#ifndef GALOIS_H
#define GALOIS_H

#include <type_traits>
#include <limits>
#include <array>
#include <vector>

#include <cstdint>

#include <arpa/inet.h>

namespace netcodec {
	inline uint8_t host_order(uint8_t arg)
	{
		return arg;
	}

	inline uint16_t host_order(uint16_t arg)
	{
		return ntohs(arg);
	}

	inline uint8_t network_order(uint8_t arg)
	{
		return arg;
	}

	inline uint16_t network_order(uint16_t arg)
	{
		return htons(arg);
	}

	template<unsigned size> struct galois_traits;

	template<> struct galois_traits<1> {
		typedef uint8_t type;
	};

	template<> struct galois_traits<2> {
		typedef uint16_t type;
	};

	template<unsigned size> class galois {
		public:
		typedef typename galois_traits<size>::type type;

		private:
		static const unsigned num_elements = 1 << 8 * size;
		static const type ltable[];
		static const type atable[];

		type element;

		public:
		galois(type element = 0)
			: element(host_order(element))
		{
		}

		galois<size> operator+(const galois<size> &arg) const
		{
			galois<size> result;
			result.element = static_cast<type>(element ^ arg.element);

			return result;
		}

		galois<size> operator*(const galois<size> &arg) const
		{
			galois<size> result;
			if (element && arg.element) {
				int index = ltable[element] + ltable[arg.element];
				index %= num_elements - 1;
				result.element = atable[index];
			}

			return result;
		}

		galois<size> operator/(const galois<size> &arg) const
		{
			galois<size> result;
			if (element && arg.element) {
				int index = num_elements - 1 + ltable[element] - ltable[arg.element];
				index %= num_elements - 1;
				result.element = atable[index];
			}

			return result;
		}

		galois<size> &operator+=(const galois<size> &arg)
		{
			*this = *this + arg;

			return *this;
		}

		galois<size> &operator-=(const galois<size> &arg)
		{
			*this = *this - arg;

			return *this;
		}

		galois<size> &operator/=(const galois<size> &arg)
		{
			*this = *this / arg;

			return *this;
		}

		bool operator==(const galois<size> &arg) const
		{
			return element == arg.element;
		}

		operator type()
		{
			return network_order(element);
		}
	};

	template<unsigned size> galois<size> operator-(const galois<size> &arg1,
			const galois<size> &arg2)
	{
		return arg1 + arg2;
	}

	template<unsigned size> bool operator!=(const galois<size> &arg1, const galois<size> &arg2)
	{
		return !(arg1 == arg2);
	}

	template<unsigned size> std::vector<typename galois<size>::type> encode(
			const std::vector<typename galois<size>::type> &coefficients,
			const std::vector<std::vector<typename galois<size>::type>> &data);

	template<unsigned size> unsigned decode( const std::vector<typename galois<size>::type> &,
			const std::vector<typename galois<size>::type> &,
			std::vector<std::vector<galois<size>>> &, unsigned);

	template<unsigned size> void back_substitute(std::vector<std::vector<galois<size>>> &);
}

#endif
