#include "get_file.h"

#include <p2pvdn/socket.h>
#include <p2pvdn/packet.h>

#include <galois.h>

#include <fstream>
#include <algorithm>

#include <poll.h>

namespace stream {
	static p2pvdn::socket start_stream(const p2pvdn::node &peer, const p2pvdn::hash &hash)
	{
		struct chord_node node;
		node.address = peer.get_address();
		node.port = peer.get_port();
		p2pvdn::socket socket(chord_oob(&node));
		socket << p2pvdn::command::stream << hash;
		return socket;
	}

	template<unsigned coefficient_size> class buffer {
		typedef typename p2pvdn::packet_type<coefficient_size>::value type;

		unsigned data_read;
		uint32_t chunk_num;
		std::vector<type> coefficients;
		std::vector<type> data;

		public:
		buffer() = default;

		buffer(unsigned num_coefficients, unsigned data_size)
			: data_read(), chunk_num(), coefficients(num_coefficients), data(data_size)
		{
		}

		bool fill(const p2pvdn::socket &socket, unsigned num_chunks)
		{
			if (data_read < sizeof(chunk_num)) {
				data_read += socket.recv(reinterpret_cast<char *>(&chunk_num) + data_read,
					sizeof(chunk_num) - data_read);

				if (data_read == sizeof(chunk_num) && get_chunk_num() == num_chunks)
					return true;
			}

			unsigned offset = sizeof(chunk_num);
			if (data_read >= offset && data_read < offset + coefficient_size * coefficients.size())
				data_read += socket.recv(
						reinterpret_cast<char *>(&coefficients[0]) + data_read - offset,
						coefficient_size * coefficients.size() - data_read + offset);

			offset += coefficient_size * coefficients.size();
			if (data_read >= offset)
				data_read += socket.recv(reinterpret_cast<char *>(&data[0]) + data_read - offset,
						coefficient_size * data.size() - data_read + offset);

			if (data_read == offset + coefficient_size * data.size())
				return true;

			return false;
		}

		unsigned get_chunk_num() const
		{
			return ntohl(chunk_num);
		}

		const std::vector<type> &get_coefficients() const
		{
			return coefficients;
		}

		const std::vector<type> &get_data() const
		{
			return data;
		}

		void empty()
		{
			data_read = 0;
		}
	};
}

template<unsigned coefficient_size> void stream::stream(get_file::ctx &ctx)
{
	std::vector<std::vector<netcodec::galois<coefficient_size>>> packets(ctx.file_opts.num_slices,
		std::vector<netcodec::galois<coefficient_size>>(
			ctx.file_opts.num_slices + ctx.file_opts.packet_elements()));

	std::vector<typename netcodec::galois<coefficient_size>::type>
		write_buffer(ctx.file_opts.packet_elements());

	std::vector<p2pvdn::socket> sockets(ctx.peers.size());
	std::vector<buffer<coefficient_size>> buffers {ctx.peers.size()};
	std::vector<pollfd> pollfds {sockets.size()};

	unsigned rank = 0;

	unsigned chunk_num = 0;
	auto socket = sockets.begin();
	auto buffer = buffers.begin();
	auto pollfd = pollfds.begin();
	for (auto peer = ctx.peers.begin(); peer != ctx.peers.end();
			peer++, socket++, buffer++, pollfd++) {

		*socket = start_stream(peer->first, ctx.hash);
		*socket << chunk_num;
		*buffer = {ctx.file_opts.num_slices, ctx.file_opts.packet_elements()};

		*pollfd = {socket->value(), POLLIN, 0};
	}

	unsigned data_rem = ctx.file_opts.filesize;
	unsigned finished_peers = 0;
	while (finished_peers != sockets.size()) {
		poll(&pollfds[0], pollfds.size(), -1);
		socket = sockets.begin();
		buffer = buffers.begin();
		for (pollfd = pollfds.begin(); pollfd != pollfds.end(); pollfd++, socket++, buffer++)
			if (pollfd->revents | POLLIN)
				while (buffer->get_chunk_num() != ctx.file_opts.num_chunks() &&
						buffer->fill(*socket, ctx.file_opts.num_chunks())) {

					if (buffer->get_chunk_num() == ctx.file_opts.num_chunks())
						finished_peers++;

					else if (buffer->get_chunk_num() == chunk_num) {
						rank = netcodec::decode<coefficient_size>(buffer->get_coefficients(),
								buffer->get_data(), packets, rank);

						if (rank == packets.size()) {
							netcodec::back_substitute<coefficient_size>(packets);
							for (auto packet = packets.begin(); packet != packets.end();
									packet++) {

								std::copy(packet->begin() + packets.size(), packet->end(),
										write_buffer.begin());

								if (data_rem > ctx.file_opts.packet_size) {
									ctx.output->write(reinterpret_cast<char *>(&write_buffer[0]),
											ctx.file_opts.packet_size);

									data_rem -= ctx.file_opts.packet_size;
								}
								else {
									ctx.output->write(reinterpret_cast<char *>(&write_buffer[0]),
											data_rem);

									data_rem = 0;
								}
							}
							rank = 0;
							chunk_num++;
							for (auto sock = sockets.begin(); sock != sockets.end(); sock++)
								*sock << chunk_num;

							if (chunk_num == ctx.file_opts.num_chunks())
								for (auto sock = sockets.begin(); sock != sockets.end(); sock++)
									*sock << p2pvdn::command::close;
						}
					}
					buffer->empty();
				}
	}
}

template void stream::stream<1>(get_file::ctx &ctx);
template void stream::stream<2>(get_file::ctx &ctx);
