#include "socket.h"

#include <sys/types.h>
#include <sys/socket.h>

void p2pvdn::socket::recvall(void *data, size_t size) const
{
	while (size) {
		ssize_t data_read = ::recv(socketfd, data, size, MSG_WAITALL);

		if (data_read == -1)
			switch (errno) {
			case EAGAIN:
#if EWOULDBLOCK != EAGAIN
			case EWOULDBLOCK:
#endif
			case EBADF:
			case ENOTSOCK:
				throw invalid_socket();

			case ECONNREFUSED:
			case ENOTCONN:
				throw connection_closed();

			case EFAULT:
			case EINVAL:
				throw std::invalid_argument("invalid address");

			case EINTR:
				break;

			default:
				throw network_error();
			}
		else if (!data_read)
			throw connection_closed();
		else {
			size -= data_read;
			data = static_cast<char *>(data) + data_read;
		}
	}
}

bool p2pvdn::socket::tryrecv(void *data, size_t size) const
{
	while (size) {
		ssize_t data_read = ::recv(socketfd, data, size, MSG_DONTWAIT);

		if (data_read == -1)
			switch (errno) {
			case EAGAIN:
#if EWOULDBLOCK != EAGAIN
			case EWOULDBLOCK:
#endif
				return false;

			case EBADF:
			case ENOTSOCK:
				throw invalid_socket();

			case ECONNREFUSED:
			case ENOTCONN:
				throw connection_closed();

			case EFAULT:
			case EINVAL:
				throw std::invalid_argument("invalid address");

			case EINTR:
				break;

			default:
				throw network_error();
			}
		else if (!data_read)
			throw connection_closed();
		else {
			size -= data_read;
			data = static_cast<char *>(data) + data_read;
			if (size)
				this->recvall(data, size);

			return true;
		}
	}

	return true;
}

unsigned p2pvdn::socket::recv(void *data, size_t size) const
{
	unsigned data_read = 0;
	while (size) {
		ssize_t return_value = ::recv(socketfd, data, size, MSG_DONTWAIT);

		if (return_value == -1)
			switch (errno) {
				case EAGAIN:
#if EWOULDBLOCK != EAGAIN
				case EWOULDBLOCK:
#endif
					return data_read;

				case EBADF:
				case ENOTSOCK:
					throw invalid_socket();

				case ECONNREFUSED:
				case ENOTCONN:
					throw connection_closed();

				case EFAULT:
				case EINVAL:
					throw std::invalid_argument("invalid address");

				case EINTR:
					break;

				default:
					throw network_error();
			}

		else if (!return_value)
			throw connection_closed();

		else {
			data_read += return_value;
			data = static_cast<char *>(data) + return_value;
			size -= return_value;
		}
	}

	return data_read;
}

void p2pvdn::socket::sendall(const void *data, size_t size) const
{
	while (size) {
		ssize_t data_wrote = send(socketfd, data, size, MSG_NOSIGNAL);

		if (data_wrote == -1)
			switch (errno) {
			case EACCES:
			case EAGAIN:
#if EWOULDBLOCK != EAGAIN
			case EWOULDBLOCK:
#endif
			case EBADF:
			case EDESTADDRREQ:
			case EISCONN:
			case EMSGSIZE:
			case ENOBUFS:
			case ENOMEM:
			case ENOTSOCK:
			case EOPNOTSUPP:
				throw invalid_socket();
			case ECONNRESET:
			case ENOTCONN:
			case EPIPE:
				throw connection_closed();
			case EFAULT:
			case EINVAL:
				throw std::invalid_argument("invalid address");
			case EINTR:
				break;
			default:
				throw network_error();
			}
		else if (!data_wrote)
			throw connection_closed();
		else {
			size -= data_wrote;
			data = static_cast<const char *>(data) + data_wrote;
		}
	}
}

void p2pvdn::socket::close()
{
	if (socketfd == -1)
		return;

	while(true) {
		int retval = ::close(socketfd);

		if (retval == -1)
			switch (errno) {
			case EBADF:
				throw invalid_socket();
			case EINTR:
				break;
			case EIO:
			default:
				throw network_error();
			}
		else
			break;
	}

	socketfd = -1;
}
