#define _GLIBCXX_USE_NANOSLEEP

#include "server.h"
#include "file_data.h"
#include "peer_data.h"
#include "vdnd.h"

#include <p2pvdn/socket.h>
#include <p2pvdn/file_opts.h>
#include <p2pvdn/hash.h>

#include <algorithm>
#include <chrono>
#include <thread>

namespace vdnd {
	static void serve_stream(const p2pvdn::socket &socket, const ctx &ctx)
	{
		auto time = std::chrono::system_clock::now();		
		auto file_data = ctx.files.get({socket});
		unsigned chunk_num = -1, prev_chunk = -1;
		while (chunk_num != file_data->num_chunks()) {
			while(tryrecv(socket, chunk_num) && chunk_num != file_data->num_chunks());
			if (chunk_num == prev_chunk)
				socket >> chunk_num;

			if (chunk_num == file_data->num_chunks())
				break;

			prev_chunk = chunk_num;

			decltype(file_data->get_chunk(chunk_num).get_packets()) packets;
			try {
				packets = file_data->get_chunk(chunk_num).get_packets();
			} catch (std::out_of_range) {
				std::clog << "error: invalid chunk requested" << std::endl;
				break;
			}

			std::random_shuffle(packets.begin(), packets.end());
			for (auto packet = packets.begin(); packet != packets.end(); packet++) {
				socket << chunk_num;
				socket << **packet;
				std::this_thread::sleep_until(time += ctx.sleep_time);
				if (tryrecv(socket, chunk_num))
					break;
			}
		}
		socket << chunk_num;
	}

	static void serve_file_opts(const p2pvdn::socket &socket, const ctx &ctx)
	{
		auto file_data = ctx.files.get({socket});
		socket << file_data->get_file_opts();
	}

	static void serve_disperse(const p2pvdn::socket &socket, ctx &ctx)
	{
		p2pvdn::hash hash(socket);
		p2pvdn::file_opts file_opts(socket);
		auto file_data = ctx.files.add(std::move(hash), std::move(file_opts));
		socket >> *file_data;
	}

	static void serve_send_peer_list(const p2pvdn::socket &socket, const ctx &ctx)
	{
		auto peer_data = ctx.peers.get({socket});
		socket << *peer_data;
	}

	static void serve_get_peer_list(const p2pvdn::socket &socket, ctx &ctx)
	{
		auto peer_data = ctx.peers.add({socket});
		socket >> *peer_data;
	}
}

int vdnd::server(int socketfd, void *arg)
{
	ctx *ctx = static_cast<vdnd::ctx *>(arg);
	const p2pvdn::socket socket(dup(socketfd));

	try {
		while (true) {
			switch (p2pvdn::command(socket)) {
				case p2pvdn::command::disperse:
					serve_disperse(socket, *ctx);
					break;
				case p2pvdn::command::key_info:
					serve_file_opts(socket, *ctx);
					break;
				case p2pvdn::command::get_peer_list:
					serve_get_peer_list(socket, *ctx);
					break;
				case p2pvdn::command::send_peer_list:
					serve_send_peer_list(socket, *ctx);
					break;
				case p2pvdn::command::stream:
					serve_stream(socket, *ctx);
					break;
				case p2pvdn::command::close:
					return 0;
				default:
					throw (std::exception());
			}
		}
	}
	catch (p2pvdn::network_error) {
		std::clog << "network error" << std::endl;
	}

	return 0;
}
